package comp2931.time;


import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.EventQueue;
import java.awt.Font;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import javax.swing.BorderFactory;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.border.Border;


/**
 * An alarm clock with a graphical user interface.
 *
 * The GUI flashes when the specified alarm time is reached.  Clicking on
 * the display will shut off the alarm.
 *
 * The alarm time can be specified on the command line in hh:mm:ss form,
 * or as a delay in seconds.  The latter must be an integer preceded
 * by a '+'.
 *
 * @author Nick Efford
 * @version 2.0 (2013-01-07)
 *
 * @see Clock
 */
public class AlarmClock extends Clock {

  private Time alarmTime;
  private boolean enabled;
  private JLabel display;


  /**
   * Creates an AlarmClock without setting the alarm.
   */
  public AlarmClock() {
    disableAlarm();
    createGui();
  }


  /**
   * Creates an AlarmClock, setting the alarm to ring at
   * the specified time.
   *
   * @param time Time object representing start time for the alarm
   */
  public AlarmClock(Time time) {
    setAlarmTime(time);
    enableAlarm();
    createGui();
  }


  /**
   * Sets the time at which this AlarmClock's alarm should start ringing
   * (and, by implication, stop ringing).  Note that this method does
   * not activate the alarm; {@link #enableAlarm} should be called to do
   * that.
   *
   * @param time Start time for the alarm
   */
  public void setAlarmTime(Time time) {
    alarmTime = time;
  }


  /**
   * Enables this AlarmClock's alarm.
   *
   * @throws UnsupportedOperationException if {@link #setAlarmTime} has
   *  not been called on this AlarmClock
   */
  public void enableAlarm() {
    if (alarmTime == null) {
      throw new UnsupportedOperationException("no alarm time set!");
    }
    enabled = true;
  }


  /**
   * Disables this AlarmClock's alarm.
   */
  public void disableAlarm() {
    alarmTime = null;
    enabled = false;
    display.setBackground(Color.lightGray);
  }


  /**
   * @return True if alarm is enabled, false otherwise
   */
  public boolean isEnabled() {
    return enabled;
  }


  /**
   * Updates current time and gives a visual indication that the alarm
   * is ringing, if necessary, by toggling the window's background colour.
   */
  @Override public void run() {
    time.tick();
    display.setText(time.toString());

    if (isEnabled() && time.compareTo(alarmTime) >= 0) {
      toggleBackground();
    }
  }


  /**
   * Simple program to create an AlarmClock, set to ring at a
   * specified time, and start it ticking.
   *
   * @param args Command-line arguments: the first is the desired
   *   alarm time as a string in hh:mm:ss format or as an offset
   *   in seconds, preceded by a '+' sign.
   */
  public static void main(String[] args) {
    if (args.length != 1) {
      System.err.println(
       "Error: time/offset required as a command line argument");
      System.exit(1);
    }

    final Time alarmTime;

    if (args[0].startsWith("+")) {
      int offset = Integer.parseInt(args[0].substring(1));
      alarmTime = new Time().plus(offset);
    }
    else {
      alarmTime = new Time(args[0]);
    }

    EventQueue.invokeLater(new Runnable() {
      public void run()
      {
        Clock myClock = new AlarmClock(alarmTime);
        myClock.start();
      }
    });
  }


  // Creates graphical user interface

  private void createGui() {
    display = new JLabel();
    display.setOpaque(true);
    display.setBackground(Color.lightGray);
    display.setForeground(Color.black);
    display.setFont(new Font("Monospaced", Font.BOLD, 64));
    display.setHorizontalAlignment(JLabel.CENTER);

    Border border = BorderFactory.createEmptyBorder(10, 15, 10, 15);
    display.setBorder(border);

    display.setText(time.toString());
    String tooltip = String.format("Alarm set for %s", alarmTime);
    display.setToolTipText(tooltip);

    display.addMouseListener(
      new MouseAdapter() {
        public void mouseReleased(MouseEvent event)
        {
          disableAlarm();
        }
      }
    );

    JFrame frame = new JFrame("AlarmClock");
    frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    frame.add(display, BorderLayout.CENTER);
    frame.pack();
    frame.setVisible(true);
  }


  // Simulates alarm by toggling background colour

  private void toggleBackground() {
    if (! display.getBackground().equals(Color.yellow)) {
      display.setBackground(Color.yellow);
    }
    else {
      display.setBackground(Color.lightGray);
    }
  }
}
