package comp2931.time;


import java.util.Calendar;


/**
 * Simple representation of a time of day, using the 24-hour clock.
 *
 * <p>Time objects can be created to represent the current time or a time
 * specified by the user in hours, minutes and seconds.</p>
 *
 * <p>Calling the {@link #tick()} method will advance the time by 1 second.
 * Calling {@link #plus(int)} will yield a new Time object advanced by the
 * specified number of seconds relative to the object on which the method
 * was called.</p>
 *
 * @author Nick Efford
 * @version 1.2 (2013-01-07)
 */
public class Time implements Comparable<Time> {

  // Class constants (shared by all Time objects)

  private static final int SECONDS_PER_MINUTE = 60;
  private static final int MINUTES_PER_HOUR = 60;
  private static final int SECONDS_PER_HOUR = 3600;
  private static final int HOURS_PER_DAY = 24;

  // Instance variables

  private int hours;
  private int minutes;
  private int seconds;


  /**
   * Creates a Time object representing the current time.
   */
  public Time() {
    Calendar now = Calendar.getInstance();

    hours = now.get(Calendar.HOUR_OF_DAY);
    minutes = now.get(Calendar.MINUTE);
    seconds = now.get(Calendar.SECOND);
  }


  /**
   * Creates a Time object representing a user-specified time.
   *
   * <p>Note that the hours component must be specified using the
   * 24-hour clock.</p>
   *
   * @param h Hours
   * @param m Minutes
   * @param s Seconds
   *
   * @throws IllegalArgumentException If hours, minutes or seconds are invalid
   */
  public Time(int h, int m, int s) {
    set(h, m, s);
  }


  /**
   * Creates a Time object representing a user-specified time.
   *
   * <p>The time is supplied as a string containing two integer values
   * representing hours and minutes or three integer values representing
   * hours, minutes and seconds, hours being specified using the 24-hour
   * clock in either case. The values must be separated from each other by
   * a colon.</p>
   *
   * @param timeString String containing hours, minutes and seconds
   * @throws IllegalArgumentException If hours, minutes or seconds are invalid
   */
  public Time(String timeString) {
    String[] parts = timeString.split(":");
    if (parts.length < 2 || parts.length > 3) {
      throw new IllegalArgumentException("ill-formed time string");
    }

    int h = Integer.parseInt(parts[0]);
    int m = Integer.parseInt(parts[1]);

    int s = 0;
    if (parts.length == 3) {
      s = Integer.parseInt(parts[2]);
    }

    set(h, m, s);
  }


  /**
   * @return Hours component of this time
   */
  public int getHours() {
    return hours;
  }


  /**
   * @return Minutes component of this time
   */
  public int getMinutes() {
    return minutes;
  }


  /**
   * @return Seconds component of this time
   */
  public int getSeconds() {
    return seconds;
  }


  /**
   * Formats this time as a string containing hours, minutes and seconds.
   *
   * Leading zeroes are added to values if necessary and components are
   * separated from each other by a colon.
   *
   * @return String representation of this time
   */
  @Override
  public String toString() {
    return String.format(
      "%02d:%02d:%02d", getHours(), getMinutes(), getSeconds());
  }


  /**
   * Tests whether this time is equal to another.
   *
   * <p>The two objects are considered equal if both are instances of
   * the Time class <em>and</em> both represent exactly the same
   * time of day.</p>
   *
   * @return True if this Time object is equal to the other, false otherwise
   */
  @Override
  public boolean equals(Object other) {
    if (other == this) {
      // 'other' is same object as this one!
      return true;
    }
    else if (! (other instanceof Time)) {
      // 'other' is not a Time object
      return false;
    }
    else {
      // Compare fields
      Time otherTime = (Time) other;
      return getHours() == otherTime.getHours()
          && getMinutes() == otherTime.getMinutes()
          && getSeconds() == otherTime.getSeconds();
    }
  }


  /**
   * Computes a hash code for this time.
   *
   * <p>This allows Time objects to work correctly in hash-based containers
   * such as {@link java.util.HashSet} and {@link java.util.HashMap}.</p>
   *
   * <p>The main requirements for a hash code are that equal objects should
   * have equal hash codes and that the hash codes for a set of unequal
   * objects should, if possible, be distributed uniformly across the range
   * of possible hash code values.</p>
   *
   * @return Hash code for this time
   */
  @Override public int hashCode() {
    // Limits on the values of hours, minutes and seconds allow us
    // to use the inSeconds method as a perfect hash; see Josh Bloch's
    // "Effective Java" for a more general hashCode recipe

    return this.inSeconds();
  }


  /**
   * Compares this time with another.
   *
   * @param other Time object to be compared with this one
   * @return Negative integer if this time is earlier than the other,
   *   positive integer if this time is later than the other, zero
   *   if both are the same time
   */
  public int compareTo(Time other) {
    int diff = hours - other.hours;

    if (diff == 0) {
      diff = minutes - other.minutes;
      if (diff == 0) {
        diff = seconds - other.seconds;
      }
    }

    return diff;
  }


  /**
   * Expresses this time as a single integer value - the number of seconds
   * since midnight.
   *
   * @return Number of seconds
   */
  public int inSeconds() {
    int hourSec = SECONDS_PER_HOUR * getHours();
    int minSec = SECONDS_PER_MINUTE * getMinutes();
    return hourSec + minSec + getSeconds();
  }


  /**
   * Creates a Time object advanced in time relative to this one by
   * a user-specified number of seconds.
   *
   * @param sec Number of seconds to advance
   * @return Time object advanced in time relative to this one
   * @throws IllegalArgumentException If number of seconds is negative
   */
  public Time plus(int sec) {
    if (sec < 0) {
      throw new IllegalArgumentException("invalid number of seconds");
    }
    else {
      int future = this.inSeconds() + sec;
      int h = (future / SECONDS_PER_HOUR) % HOURS_PER_DAY;
      int remainder = future % SECONDS_PER_HOUR;
      int m = remainder / SECONDS_PER_MINUTE;
      int s = remainder % SECONDS_PER_MINUTE;
      return new Time(h, m, s);
    }
  }


  /**
   * Advances this time by one second.
   */
  public void tick() {
    seconds = (seconds + 1) % SECONDS_PER_MINUTE;
    if (seconds == 0) {
      minutes = (minutes + 1) % MINUTES_PER_HOUR;
      if (minutes == 0)
        hours = (hours + 1) % HOURS_PER_DAY;
    }
  }


  // Helper method to assign values to instance variables
  // (not part of public interface!)

  private void set(int h, int m, int s) {
     if (h < 0 || h >= HOURS_PER_DAY) {
      throw new IllegalArgumentException("hours out of range");
    }
    else if (m < 0 || m >= MINUTES_PER_HOUR) {
      throw new IllegalArgumentException("minutes out of range");
    }
    else if (s < 0 || s >= SECONDS_PER_MINUTE) {
      throw new IllegalArgumentException("seconds out of range");
    }
    else {
      hours = h;
      minutes = m;
      seconds = s;
    }
  }
}
